import numpy as np
from .Activation_Functions import *


class Layer:

    dz = []
    dw = []
    db = []

    def __init__(self, inputs, neurons, activation_function, last_layer, first_layer, id_num):
        self.weights = np.random.uniform(low=-0.5, high=0.5, size=(neurons, inputs))
        self.bias = np.ones(neurons)
        self.z = np.zeros(neurons)
        self.dz = np.zeros(neurons)
        self.output = np.zeros(neurons)
        self.activation_function = activation_function
        self.last_layer = last_layer
        self.first_layer = first_layer
        self.neurons = neurons
        self.id_num = id_num

    def forward_propagation(self, input_values):
        self.z = self.weights.dot(input_values) + self.bias
        self.output = np.array([self.activation_function(self.z, x) for x in self.z])
        return self.output

    def back_propagation(self, training_output, training_input, prev_layer=None, next_layer=None):
        if self.last_layer and self.first_layer:
            pass
        elif self.last_layer:
            self.dz = self.output - training_output
            temp = self.dz.reshape((-1, 1))
            self.dw = temp * prev_layer.output
            self.db = self.dz
        else:
            temp1 = next_layer.weights.T
            temp2 = next_layer.dz.reshape((-1, 1))
            self.dz = temp1.dot(temp2).T * np.array([relu_deriv(x) for x in self.z])
            if self.first_layer:
                self.dw = self.dz.reshape((-1, 1)) * training_input
            else:
                self.dw = self.dz.reshape((-1, 1)) * prev_layer.output
            self.db = self.dz

    def update_values(self, learning_rate):
        self.weights = self.weights - learning_rate * self.dw
        self.bias = self.bias - learning_rate * self.db
        self.bias = self.bias.reshape(self.neurons, )

    def save_to_file(self):
        file_name_weights = "savedWages/wages/w_" + str(self.id_num) + ".csv"
        file_name_biases = "savedWages/biases/b_" + str(self.id_num) + ".csv"
        np.savetxt(file_name_weights, self.weights, delimiter=',')
        np.savetxt(file_name_biases, self.bias, delimiter=',')

    def load_from_file(self):
        file_name_weights = "savedWages/wages/w_" + str(self.id_num) + ".csv"
        file_name_biases = "savedWages/biases/b_" + str(self.id_num) + ".csv"
        self.weights = np.loadtxt(file_name_weights, delimiter=',')
        self.bias = np.loadtxt(file_name_biases, delimiter=',')



import numpy as np
import scipy

def relu(array=None, x=None):
    return 0 if x < 0 else x

def relu_deriv(x):
    return 0 if x < 0 else 1

def softmax(array=None, x=None):
    return scipy.special.softmax(array)[np.where(array == x)[0][0]]



from .Layer import Layer
import numpy as np


class NN:


    def __init__(self, layers, learning_rate):
        self.learning_rate = learning_rate
        self.layers = []
        temp_layer = 0
        for layer in range(len(layers)):
            self.add_layer(inputs=layers[temp_layer][0], 
                           neurons=layers[layer][0], 
                           activation_function=layers[layer][1], 
                           last_layer=True if layer == len(layers) - 1 else False,
                           first_layer=True if temp_layer == 0 else False,
                           id=layer)
            temp_layer = layer

    def add_layer(self, inputs, neurons, activation_function, last_layer, first_layer, id):
        self.layers.append(Layer(inputs, 
                                 neurons, 
                                 activation_function, 
                                 last_layer, 
                                 first_layer, 
                                 id))

    def predict(self, num_of_train):
        input_for_next_layer = self.full_data_input[num_of_train]
        for layer_num in range(len(self.layers)):
            input_for_next_layer = self.layers[layer_num].forward_propagation(input_for_next_layer)
        return np.argmax(input_for_next_layer)

    def full_backward_pass(self, num_of_training):
        number_of_layers = len(self.layers) - 1 if len(self.layers) - 1 != 0 else len(self.layers)
        for layer_num in range(number_of_layers, -1, -1):
            prev_layer = self.layers[layer_num - 1] if layer_num - 1 >= 0 else None
            next_layer = self.layers[layer_num + 1] if layer_num + 1 < len(self.layers) else None
            self.layers[layer_num].back_propagation(self.full_data_output[num_of_training],
                                                    self.full_data_input[num_of_training], 
                                                    prev_layer, 
                                                    next_layer)

    def load_data_with_output(self, data):
        self.full_data_input = []
        self.full_data_output = []
        max_number_in_input_data = np.amax(np.abs(np.array(data)))
        for i in data:
            self.full_data_input.append(np.array(i)[1:] / max_number_in_input_data)
            self.full_data_output.append(self.create_output(i[0], 
                                         len(i) - 1))

    @staticmethod
    def create_output(etiquette, number_of_possible_outputs):
        data_output = []
        for i in range(0, number_of_possible_outputs):
            data_output.append(1 if i == etiquette else 0)
        return data_output

    def load_data(self, data):
        self.full_data_input = []
        max_number_in_input_data = np.amax(np.abs(np.array(data)))
        for i in data:
            self.full_data_input.append(np.array(i) / max_number_in_input_data)

    def learn(self, max_epoch_number):
        accuracy, prev_accuracy = 0, 0
        for i in range(max_epoch_number):
            temp = 0
            for j in range(0, len(self.full_data_input)):
                prediction = self.predict(j)
                temp += self.check_if_same_result(prediction, np.argmax(self.full_data_output[j]))
                self.full_backward_pass(j)
                self.update_all()
            break_flag, prev_accuracy = self.calculate_accuracy(temp, prev_accuracy, accuracy, i)
            if break_flag: break

    @staticmethod
    def check_if_same_result(prediction, etiquette):
        return 1 if prediction == etiquette else 0

    def calculate_accuracy(self, temp, prev_accuracy, accuracy, epoch):
        accuracy = temp/len(self.full_data_input)
        if accuracy == prev_accuracy and accuracy > 0.999:
            return True, accuracy
        print("Accuracy: " + str(temp/len(self.full_data_input)) + " | Epoch: " + str(epoch + 1))
        return False, accuracy

    def update_all(self):
        for layer_num in range(0, len(self.layers) - 1):
            self.layers[layer_num].update_values(self.learning_rate)

    def save_variables(self):
        for layer in self.layers:
            layer.save_to_file()

    def load_variables(self):
        for layer in self.layers:
            layer.load_from_file()



from setuptools import setup, find_packages

VERSION = '1.171'
DESCRIPTION = 'Neural network framework'
LONG_DESCRIPTION = 'Neural network framework for learning purposes'

setup(
    name="VMNeuralNetwork",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="Var Moon",
    author_email="bbala1351998@gmail.com",
    license='MIT',
    packages=find_packages(),
    install_requires=['numpy', 
                      'scipy'],
    keywords='conversion',
    classifiers= [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
    ]
)
